Legend:
Page
Library
Module
Module type
Parameter
Class
Class type
Page
Library
Module
Module type
Parameter
Class
Class type
Tiny_httpd.Byte_streamSourceMake a buffered stream from the given channel.
Same as of_chan but the close method will never fail.
A stream that just returns the slice of bytes starting from i and of length len.
Write the stream to the channel.
Open a file with given name, and obtain an input stream on its content. When the function returns, the stream (and file) are closed.